/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.impl;

import org.xmloperator.lambda2.net.model.BinaryCell;
import org.xmloperator.lambda2.net.model.Cell;
import org.xmloperator.lambda2.net.model.Port;
import org.xmloperator.lambda2.net.model.Stack;

/**
 * BinaryCell implementation.
 */
public final class BinaryCellImpl extends CellImpl implements BinaryCell {
	private static final boolean ASSERTION = Test.ASSERTION;

	final PortImpl port1 = new PortImpl(this, 1);
	final PortImpl port2 = new PortImpl(this, 2);
	boolean isDuplicator;

	public BinaryCellImpl(boolean isDuplicator) {
		this.isDuplicator = isDuplicator;
	}

	//////////////////////
	// Cell implementation
	//////////////////////

	public final int getSymbol() {
		return this.isDuplicator ? SYMBOL_DUPLICATOR : SYMBOL_CONSTRUCTOR;
	}

	public int getSymbolArity() {
		return 2;
	}
	
	public final boolean isBinaryCell() {
		return true;
	}

	public Port getPort(int portNumber) {
		switch (portNumber) {
			case 0 :
				return this;
			case 1 :
				return this.port1;
			case 2 :
				return this.port2;
			default :
				throw new IllegalArgumentException();
		}
	}

	public void collectPorts(Stack portStack) {
		super.collectPorts(portStack);
		portStack.push(this.port1);
		portStack.push(this.port2);
	}

	public Cell cloneCell() {
		return new BinaryCellImpl(this.isDuplicator);
	}

	public final int interact() {
		PortImpl connectedPort = this.connectedPort;
		if (connectedPort.portNumber != 0)
			throw new IllegalStateException("Not an active pair");
		CellImpl cell = connectedPort.cell;
		switch (cell.getSymbol()) {
			case SYMBOL_ERASER :
				this.port1.connectedPort.connectTo(cell);
				this.port2.connectedPort.connectTo(new EraserCellImpl());
				return this.isDuplicator
						? INTERACTION_ERASING_DUPLICATOR
						: INTERACTION_ERASING_CONSTRUCTOR;
			case SYMBOL_DUPLICATOR :
			case SYMBOL_CONSTRUCTOR :
				return this.interactWithBinaryCell((BinaryCellImpl) cell);
			default :
				return cell.interact();
		}
	}

	private final int interactWithBinaryCell(BinaryCellImpl binaryCell) {
		PortImpl connectedPort11 = this.port1.connectedPort;
		PortImpl connectedPort12 = this.port2.connectedPort;
		PortImpl connectedPort21 = binaryCell.port1.connectedPort;
		PortImpl connectedPort22 = binaryCell.port2.connectedPort;
		if (ASSERTION
				&& (connectedPort11 == this.port2 || connectedPort21 == binaryCell.port2))
			throw new IllegalStateException("Illegal interconnection");
		if (binaryCell.isDuplicator != this.isDuplicator) {
			BinaryCellImpl duplicator = this.isDuplicator ? this : binaryCell;
			PortImpl connectedPort1 = duplicator.port1.connectedPort;
			PortImpl connectedPort2 = duplicator.port2.connectedPort;
			if (connectedPort1.portNumber == 0
					&& connectedPort2.portNumber == 0
					&& connectedPort1.cell.getSymbol() == SYMBOL_CONSTRUCTOR
					&& connectedPort2.cell.getSymbol() == SYMBOL_CONSTRUCTOR) {
				// C-DC1C2 -> C-CD1D2 with D12 and D21 permuted
				// 1) Change C-D to C-C and annihilate C-C
				connectedPort11.connectTo(connectedPort22);
				connectedPort12.connectTo(connectedPort21);
				// 2) Change C1C2 to D1D2
				BinaryCellImpl cell1 = (BinaryCellImpl) connectedPort1.cell;
				BinaryCellImpl cell2 = (BinaryCellImpl) connectedPort2.cell;
				cell1.isDuplicator = true;
				cell2.isDuplicator = true;
				// 3) Permut D12 and D21
				connectedPort12 = cell1.port2.connectedPort;
				connectedPort21 = cell2.port1.connectedPort;
				cell1.port2.connectTo(connectedPort21);
				cell2.port1.connectTo(connectedPort12);
				return INTERACTION_CONSTRUCTOR_DUPLICATOR_CONTRUCTORS;
			}
			if (ASSERTION
					&& (connectedPort11.cell == binaryCell
							|| connectedPort12.cell == binaryCell
							|| connectedPort21.cell == this || connectedPort22.cell == this))
				throw new IllegalStateException("Illegal interconnection");
			BinaryCellImpl cell11 = this;
			BinaryCellImpl cell12 = new BinaryCellImpl(this.isDuplicator);
			BinaryCellImpl cell21 = binaryCell;
			BinaryCellImpl cell22 = new BinaryCellImpl(binaryCell.isDuplicator);
			connectedPort11.connectTo(cell21);
			connectedPort12.connectTo(cell22);
			connectedPort21.connectTo(cell11);
			connectedPort22.connectTo(cell12);
			cell11.port1.connectTo(cell21.port1);
			cell11.port2.connectTo(cell22.port1);
			cell12.port1.connectTo(cell21.port2);
			cell12.port2.connectTo(cell22.port2);
			return INTERACTION_CONSTRUCTOR_DUPLICATOR;
		} else if (this.isDuplicator) {
			if (connectedPort11.cell == binaryCell) {
				PortImpl oppositePort = connectedPort11 == binaryCell.port1
						? connectedPort22
						: connectedPort21;
				connectedPort12.connectTo(oppositePort);
			} else if (connectedPort12.cell == binaryCell) {
				PortImpl oppositePort = connectedPort12 == binaryCell.port1
						? connectedPort22
						: connectedPort21;
				connectedPort11.connectTo(oppositePort);
			} else {
				connectedPort11.connectTo(connectedPort21);
				connectedPort12.connectTo(connectedPort22);
			}
			return INTERACTION_ANNIHILATION_DUPLICATOR;
		} else {
			if (connectedPort11.cell == binaryCell) {
				PortImpl oppositePort = connectedPort11 == binaryCell.port1
						? connectedPort22
						: connectedPort21;
				connectedPort12.connectTo(oppositePort);
			} else if (connectedPort12.cell == binaryCell) {
				PortImpl oppositePort = connectedPort12 == binaryCell.port1
						? connectedPort22
						: connectedPort21;
				connectedPort11.connectTo(oppositePort);
			} else {
				connectedPort11.connectTo(connectedPort22);
				connectedPort12.connectTo(connectedPort21);
			}
			return INTERACTION_ANNIHILATION_CONSTRUCTOR;
		}
	}

	////////////////////////////
	// BinaryCell implementation
	////////////////////////////

	public final Port getPort1() {
		return this.port1;
	}

	public final Port getPort2() {
		return this.port2;
	}

	public final boolean isDuplicator() {
		return this.isDuplicator;
	}

	public final void changeToDuplicator() {
		this.isDuplicator = true;
	}
}