/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.impl;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.xmloperator.lambda2.net.model.Cell;
import org.xmloperator.lambda2.net.model.IntStack;
import org.xmloperator.lambda2.net.model.LambdaNet;
import org.xmloperator.lambda2.net.model.LongStack;
import org.xmloperator.lambda2.net.model.Stack;
import org.xmloperator.lambda2.net.model.TooMuchReductionStepException;
import org.xmloperator.lambda2.net.model.Port;
import org.xmloperator.lambda2.term.impl.TermImpl;
import org.xmloperator.lambda2.term.impl.VariableImpl;
import org.xmloperator.lambda2.term.model.Term;
import org.xmloperator.lambda2.term.model.Variable;

/**
 * LambdaNet implementation.
 */
public final class LambdaNetImpl implements LambdaNet {
	private static final boolean IS_FUNCTION_EXPONENT_ENABLED = true;

	final PortImpl bodySupport = new PortImpl();
	private int freeVariableCount;

	/**
	 * Constructor of a LambdaNet that represents a free variable.
	 */
	public LambdaNetImpl() {
		this.freeVariableCount = 1;
		this.bodySupport.connectTo(new PortImpl());
	}

	/**
	 * Constructor of a cloned LambdaNet.
	 * 
	 * @param bodySupport
	 *            a source LambdaNet body support.
	 */
	public LambdaNetImpl(final PortImpl bodySupport) {
		this.freeVariableCount = 0;
		Map cellMap = new HashMap();
		Stack oldPortStack = new StackImpl();
		Stack newPortStack = new StackImpl();
		oldPortStack.push(bodySupport);
		newPortStack.push(this.bodySupport);
		while (!oldPortStack.isEmpty()) {
			PortImpl oldPort = (PortImpl) oldPortStack.pop();
			PortImpl newPort = (PortImpl) newPortStack.pop();
			if (newPort.connectedPort != null)
				continue;
			PortImpl oldConnectedPort = oldPort.connectedPort;
			CellImpl oldCell = oldConnectedPort.cell;
			if (oldCell == null) {
				this.freeVariableCount++;
				newPort.connectTo(new PortImpl());
				continue;
			}
			int portNumber = oldConnectedPort.portNumber;
			CellImpl newCell = (CellImpl) cellMap.get(oldCell);
			if (newCell == null) {
				newCell = (CellImpl) oldCell.cloneCell();
				cellMap.put(oldCell, newCell);
				oldCell.collectPorts(oldPortStack);
				newCell.collectPorts(newPortStack);
			}
			newPort.connectTo(newCell.getPort(portNumber));
		}
	}

	/**
	 * Print a graph representation of this LambdaNet.
	 * 
	 * @param out
	 *            a PrintStream.
	 */
	public void printNet(PrintStream out) {
		printNet(out, this.bodySupport);
	}

	private static final void printNet(final PrintStream out,
			final PortImpl bodySupport) {
		List cellList = new ArrayList();
		List supportList = new ArrayList();
		supportList.add(bodySupport);
		Set cellSet = new HashSet();
		Stack portStack = new StackImpl();
		portStack.push(bodySupport);
		while (!portStack.isEmpty()) {
			PortImpl port = (PortImpl) portStack.pop();
			PortImpl connectedPort = port.connectedPort;
			if (connectedPort == null)
				continue;
			CellImpl cell = connectedPort.cell;
			if (cell == null) {
				supportList.add(connectedPort);
				continue;
			}
			if (!cellSet.add(cell))
				continue;
			cellList.add(cell);
			cell.collectPorts(portStack);
		}
		cellSet.clear();
		portStack.push(bodySupport);
		while (!portStack.isEmpty()) {
			PortImpl port = (PortImpl) portStack.pop();
			PortImpl connectedPort = port.connectedPort;
			if (connectedPort == null)
				continue;
			CellImpl cell = connectedPort.cell;
			if (cell == null || !cellSet.add(cell))
				continue;
			int cellNumber = 1 + cellList.indexOf(cell);
			cell.collectPorts(portStack);
			int symbol = cell.getSymbol();
			out.print("Cell(");
			out.print(cellNumber);
			out.print(":");
			out.print(Cell.SYMBOL_SYMBOLS[symbol]);
			out.print(")");
			int arity = cell.getSymbolArity();
			for (int k = 0; k <= arity; k++) {
				out.print(" ");
				port = (PortImpl) cell.getPort(k);
				connectedPort = port.connectedPort;
				if (connectedPort == null) {
					out.print("-");
					continue;
				}
				CellImpl connectedCell = connectedPort.cell;
				if (connectedCell == null) {
					out.print("S");
					out.print(supportList.indexOf(connectedPort));
					continue;
				}
				out.print(1 + cellList.indexOf(connectedCell));
				out.print(".");
				out.print(connectedPort.portNumber);
			}
			out.println();
		}
	}

	/**
	 * Print a statistic of symbols of this LambdaNet.
	 * 
	 * @param out
	 *            a PrintStream.
	 */
	public void printStat(PrintStream out) {
		printStat(out, this.bodySupport);
	}

	private static final void printStat(final PrintStream out,
			final PortImpl bodySupport) {
		int[] symbolStat = new int[Cell.SYMBOL_SYMBOLS.length];
		Arrays.fill(symbolStat, 0);
		Set cellSet = new HashSet();
		Stack portStack = new StackImpl();
		portStack.push(bodySupport);
		while (!portStack.isEmpty()) {
			PortImpl port = (PortImpl) portStack.pop();
			PortImpl connectedPort = port.connectedPort;
			if (connectedPort == null)
				continue;
			CellImpl cell = connectedPort.cell;
			if (cell == null || !cellSet.add(cell))
				continue;
			symbolStat[cell.getSymbol()]++;
			cell.collectPorts(portStack);
		}
		out.print("Symbols:");
		int total = 0;
		for (int k = 0; k < symbolStat.length; k++)
			if (symbolStat[k] != 0) {
				out.print(" [");
				out.print(Cell.SYMBOL_SYMBOLS[k]);
				out.print("]:");
				out.print(symbolStat[k]);
				total += symbolStat[k];
			}
		out.println(" [total]:" + total);
	}

	/**
	 * Returns the count of cells of this LambdaNet.
	 * 
	 * @return the count of cells of this LambdaNet.
	 */
	public int cellCount() {
		return cellCount(this.bodySupport);
	}

	private static final int cellCount(final PortImpl bodySupport) {
		Set cellSet = new HashSet();
		Stack portStack = new StackImpl();
		portStack.push(bodySupport);
		while (!portStack.isEmpty()) {
			PortImpl port = (PortImpl) portStack.pop();
			PortImpl connectedPort = port.connectedPort;
			if (connectedPort == null)
				continue;
			CellImpl cell = connectedPort.cell;
			if (cell == null || !cellSet.add(cell))
				continue;
			cell.collectPorts(portStack);
		}
		return cellSet.size();
	}

	public static void main(String[] args) {
		if (test(java.lang.System.out, true))
			java.lang.System.out.println("All is OK");
		java.lang.System.exit(0);
	}

	private static final boolean IS_VERBOSE = false;

	/**
	 * Try some LambdaNets.
	 * 
	 * @param out
	 *            the PrintStream to print results.
	 * @param isVerbose
	 *            if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(final PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			boolean isOk = true;
			{
				LambdaNetImpl t1 = new LambdaNetImpl();
				Port p1 = t1.getBodySupport().getConnectedPort();
				t1.buildAbstraction(p1); // \x1.x1
				isOk &= test(out, isVerbose, t1, "\\x1.x1");
			}
			{
				LambdaNetImpl t1 = new LambdaNetImpl();
				Port p1 = t1.getBodySupport().getConnectedPort();
				LambdaNetImpl t2 = new LambdaNetImpl();
				Port p2 = t2.getBodySupport().getConnectedPort();
				LambdaNetImpl t3 = new LambdaNetImpl();
				Port p3 = t3.getBodySupport().getConnectedPort();
				t1.buildApplication(t2);
				t3.buildApplication(t1);
				t3.buildAbstraction(new Port[]{p1, p2, p3}); // \x1.x1(x1x1)
				isOk &= test(out, isVerbose, t3, "\\x1.x1(x1x1)");
			}
			{
				LambdaNetImpl t1 = new LambdaNetImpl();
				Port p1 = t1.getBodySupport().getConnectedPort();
				LambdaNetImpl t2 = new LambdaNetImpl();
				Port p2 = t2.getBodySupport().getConnectedPort();
				LambdaNetImpl t3 = new LambdaNetImpl();
				Port p3 = t3.getBodySupport().getConnectedPort();
				t2.buildApplication(t3);
				t1.buildApplication(t2);
				t1.buildAbstraction(new Port[]{p3});
				t1.buildAbstraction(new Port[]{p1, p2}); // \x1.\x2.x1(x1x2)
				isOk &= test(out, isVerbose, t1, "\\x1.\\x2.x1(x1x2)");
				isOk &= test(out, isVerbose, t1, 2);
			}
			return isOk;
		} catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static boolean test(final PrintStream out, final boolean isVerbose,
			LambdaNet lambdaNet, String expectedText) {
		String text = lambdaNet.readBack(false, -1, null).toText();
		boolean isOk = text.equals(expectedText);
		if (isVerbose || !isOk)
			out.println("LambdaNetImpl " + (isOk ? "ok" : "ko") + " \"" + text
					+ "\" / \"" + expectedText + '\"');
		return isOk;
	}

	private static boolean test(final PrintStream out, final boolean isVerbose,
			LambdaNet lambdaNet, long expectedValue) {
		long value = lambdaNet.numeralValue(null);
		boolean isOk = value == expectedValue;
		if (isVerbose || !isOk)
			out.println("LambdaNetImpl " + (isOk ? "ok" : "ko") + " " + value
					+ " / " + expectedValue);
		return isOk;
	}

	// /////////////////////////
	// LambdaNet implementation
	// /////////////////////////

	public Port getBodySupport() {
		return this.bodySupport;
	}

	public int getFreeVariableCount() {
		return this.freeVariableCount;
	}

	private static final Port[] NO_VARIABLE_SUPPORT = new Port[]{};

	public void buildAbstraction() {
		this.buildAbstraction(NO_VARIABLE_SUPPORT);
	}

	public void buildAbstraction(Port freeVariableSupport) {
		this.buildAbstraction(new Port[]{freeVariableSupport});
	}

	public void buildAbstraction(final Port[] freeVariableSupports) {
		if (this.freeVariableCount < freeVariableSupports.length)
			throw new IllegalArgumentException();
		this.freeVariableCount -= freeVariableSupports.length;
		Port boundVariableSupport;
		if (freeVariableSupports.length == 0) {
			boundVariableSupport = new PortImpl();
			boundVariableSupport.connectTo(new EraserCellImpl());
		} else {
			boundVariableSupport = freeVariableSupports[0];
			for (int k = 1; k < freeVariableSupports.length; k++) {
				Port freeVariableSupport = freeVariableSupports[k];
				if (freeVariableSupport.getConnectedPort() == boundVariableSupport)
					throw new IllegalStateException();
				BinaryCellImpl duplicator = new BinaryCellImpl(true);
				duplicator.port1.connectTo(boundVariableSupport
						.getConnectedPort());
				duplicator.port2.connectTo(freeVariableSupport
						.getConnectedPort());
				duplicator.connectTo(boundVariableSupport);
			}
		}
		Port duplicatorPort0;
		Port duplicatorPort1;
		Port duplicatorPort2;
		BinaryCellImpl duplicator = (BinaryCellImpl) Duplicators
				.extractDuplicator(this.bodySupport);
		if (duplicator != null) {
			duplicatorPort0 = duplicator.connectedPort;
			duplicatorPort1 = duplicator.port1.connectedPort;
			duplicatorPort2 = duplicator.port2.connectedPort;
			duplicatorPort0.disconnect();
			duplicatorPort1.disconnect();
			duplicatorPort2.disconnect();
		} else {
			duplicatorPort0 = new EraserCellImpl();
			duplicatorPort1 = new EraserCellImpl();
			duplicatorPort2 = new EraserCellImpl();
		}
		Port bodyPort = this.bodySupport.connectedPort;
		Port boundVariablePort = boundVariableSupport.getConnectedPort();
		this.bodySupport.disconnect();
		boundVariableSupport.disconnect();
		PentaryMultiplexorImpl m5 = new PentaryMultiplexorImpl(true);
		m5.connectTo(this.bodySupport);
		m5.duplicatorPort1.connectTo(duplicatorPort1);
		m5.duplicatorPort2.connectTo(duplicatorPort2);
		if (boundVariablePort == this.bodySupport) {
			m5.variablePort.connectTo(m5.bodyPort);
			m5.duplicatorPort0.connectTo(duplicatorPort0);
		} else if (duplicatorPort0 == boundVariableSupport) {
			m5.duplicatorPort0.connectTo(m5.variablePort);
			m5.bodyPort.connectTo(bodyPort);
		} else {
			m5.duplicatorPort0.connectTo(duplicatorPort0);
			m5.variablePort.connectTo(boundVariablePort);
			m5.bodyPort.connectTo(bodyPort);
		}
	}

	public void buildApplication(final LambdaNet argument) {
		this.buildApplication(argument, 1);
	}

	public void buildApplication(LambdaNet argument, int functionExponent) {
		if (argument == this)
			throw new IllegalArgumentException("Cannot apply to itself");
		if (functionExponent < 1)
			throw new IllegalArgumentException("Illegal exponent");
		PortImpl functionPort = this.bodySupport.connectedPort;
		PortImpl argumentPort = ((LambdaNetImpl) argument).bodySupport.connectedPort;
		functionPort.disconnect();
		argumentPort.disconnect();
		for (int k = 1; k < functionExponent; k++) {
			BinaryCellImpl duplicator = new BinaryCellImpl(true);
			duplicator.connectTo(functionPort);
			functionPort = duplicator.port1;
			argumentPort = buildApplication(duplicator.port2, argumentPort);
		}
		argumentPort = buildApplication(functionPort, argumentPort);
		this.bodySupport.connectTo(argumentPort);
		this.freeVariableCount += argument.getFreeVariableCount();
	}

	private static final PortImpl buildApplication(final PortImpl functionPort,
			final PortImpl argumentPort) {
		PentaryMultiplexorImpl m5 = new PentaryMultiplexorImpl(false);
		BinaryCellImpl duplicator = new BinaryCellImpl(true);
		m5.connectTo(functionPort);
		m5.variablePort.connectTo(argumentPort);
		m5.duplicatorPort2.connectTo(duplicator.port2);
		m5.duplicatorPort1.connectTo(duplicator.port1);
		m5.duplicatorPort0.connectTo(duplicator);
		return m5.bodyPort;
	}

	public int reduceUntilInterfaceNormalForm(int maxStepCount,
			final int[] interactionStat) throws TooMuchReductionStepException {
		return reduceUntilInterfaceNormalForm(this.bodySupport, maxStepCount,
				interactionStat);
	}

	private static final int reduceUntilInterfaceNormalForm(
			final PortImpl bodySupport, int maxStepCount,
			final int[] interactionStat) throws TooMuchReductionStepException {
		int stepCount = 0;
		while (true) {
			PortImpl port = bodySupport.connectedPort;
			if (port.portNumber <= 0)
				return stepCount;
			do
				port = port.cell.connectedPort;
			while (port.portNumber > 0);
			if (port.portNumber != 0)
				return stepCount;
			if (stepCount == maxStepCount)
				throw new TooMuchReductionStepException();
			{
				int interactionType = port.cell.interact();
				if (interactionStat != null)
					interactionStat[interactionType]++;
			}
			stepCount++;
		}
	}

	private static final int DECODING_NODE_ABSTRACTION = 0;
	private static final int DECODING_NODE_FUNCTION = 1;
	private static final int DECODING_NODE_ARGUMENT = 2;

	public Term readBack(final boolean inNormalForm, int maxStepCount,
			final int[] interactionStat) throws TooMuchReductionStepException {
		if (this.freeVariableCount != 0)
			throw new IllegalStateException("There is free variable(s)");
		PortImpl bodySupport = new LambdaNetImpl(this.bodySupport).bodySupport;
		IntStack decodingStack = new IntStackImpl();
		List abstractionVariablesStack = new ArrayList();
		Stack functionTermStack = new StackImpl();
		LongStack sequenceCountStack = new LongStackImpl();
		Stack argumentSupportStack = new StackImpl();
		while (true) {
			Term result = null;
			long sequenceCount = 1L;
			PortImpl bodyPort = bodySupport.connectedPort;
			CellImpl bodyCell = bodyPort.cell;
			if (bodyPort.portNumber != 0
					&& (inNormalForm || !(bodyPort.portNumber == 1 && bodyCell
							.getSymbol() == Cell.SYMBOL_APPLICATOR))) {
				int stepCount = reduceUntilInterfaceNormalForm(bodySupport,
						maxStepCount, interactionStat);
				if (maxStepCount >= 0)
					maxStepCount -= stepCount;
				bodyPort = bodySupport.connectedPort;
				bodyCell = bodyPort.cell;
			}
			switch (bodyPort.portNumber) {
				case 0 :
					switch (bodyCell.getSymbol()) {
						case Cell.SYMBOL_DECODER : {
							DecoderImpl decoder = (DecoderImpl) bodyCell;
							int arity = decoder.getSymbolArity();
							if (arity == 1)
								sequenceCount = decoder.getSequenceCount();
							while (arity != 0) {
								PortImpl argumentPort = ((PortImpl) decoder
										.getPort(arity--)).connectedPort;
								PortImpl argumentSupport = new PortImpl();
								argumentPort.connectTo(argumentSupport);
								argumentSupportStack.push(argumentSupport);
								decodingStack.push(DECODING_NODE_FUNCTION);
							}
							int abstractionIndex = decoder.getIndex();
							VariableImpl variable = new VariableImpl();
							((List) abstractionVariablesStack
									.get(abstractionIndex)).add(variable);
							result = variable;
							break;
						}
						case Cell.SYMBOL_ABSTRACTOR : {
							List variableList = new ArrayList();
							int abstractionIndex = abstractionVariablesStack
									.size();
							abstractionVariablesStack.add(variableList);
							bodyPort = buildApplication(bodyPort,
									new DecoderImpl(abstractionIndex));
							bodyPort.connectTo(bodySupport);
							{
								int interactionType = bodyPort.cell.interact();
								if (interactionStat != null)
									interactionStat[interactionType]++;
							}
							decodingStack.push(DECODING_NODE_ABSTRACTION);
							continue;
						}
						default :
							throw new IllegalStateException(
									"Abstractor or decoder expected");
					}
					break;
				case 1 : {
					if (bodyCell.getSymbol() != Cell.SYMBOL_APPLICATOR)
						throw new IllegalStateException("Applicator expected");
					PentaryMultiplexorImpl m5 = (PentaryMultiplexorImpl) bodyCell;
					bodyPort = m5.connectedPort;
					bodyPort.connectTo(bodySupport);
					PortImpl argumentPort = m5.variablePort.connectedPort;
					PortImpl argumentSupport = new PortImpl();
					argumentPort.connectTo(argumentSupport);
					argumentSupportStack.push(argumentSupport);
					decodingStack.push(DECODING_NODE_FUNCTION);
					continue;
				}
				default :
					throw new IllegalStateException("Port 0 or 1 expected");
			}
			while (true) {
				if (decodingStack.isEmpty())
					return result;
				switch (decodingStack.pop()) {
					case DECODING_NODE_ABSTRACTION : {
						List variableList = (List) abstractionVariablesStack
								.remove(abstractionVariablesStack.size() - 1);
						switch (variableList.size()) {
							case 0 :
								result = result.buildAbstraction();
								break;
							case 1 :
								result = result
										.buildAbstraction((Variable) variableList
												.get(0));
								break;
							default :
								Variable[] boundVariables = new Variable[variableList
										.size()];
								variableList.toArray(boundVariables);
								result = result
										.buildAbstraction(boundVariables);
						}
						continue;
					}
					case DECODING_NODE_FUNCTION :
						functionTermStack.push(result);
						sequenceCountStack.push(sequenceCount);
						decodingStack.push(DECODING_NODE_ARGUMENT);
						bodySupport = (PortImpl) argumentSupportStack.pop();
						break;
					case DECODING_NODE_ARGUMENT : {
						TermImpl function = (TermImpl) functionTermStack.pop();
						long count = sequenceCountStack.pop();
						if (IS_FUNCTION_EXPONENT_ENABLED)
							result = function.buildApplication(result, count);
						else
							for (int k = 0; k < count; k++)
								result = function.buildApplication(result);
						continue;
					}
					default :
						throw new IllegalStateException();
				}
				break;
			}
		}
	}

	public boolean booleanValue(final int[] interactionStat) {
		if (this.freeVariableCount != 0)
			throw new IllegalStateException("There is free variable(s)");
		PortImpl bodySupport = new LambdaNetImpl(this.bodySupport).bodySupport;
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		PortImpl bodyPort = bodySupport.connectedPort;
		if (bodyPort.portNumber != 0
				|| bodyPort.cell.getSymbol() != Cell.SYMBOL_ABSTRACTOR)
			throw new IllegalStateException();
		bodySupport.connectTo(buildApplication(bodyPort, new DecoderImpl(0)));
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		bodyPort = bodySupport.connectedPort;
		if (bodyPort.portNumber != 0
				|| bodyPort.cell.getSymbol() != Cell.SYMBOL_ABSTRACTOR)
			throw new IllegalStateException();
		bodySupport.connectTo(buildApplication(bodyPort, new DecoderImpl(1)));
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		bodyPort = bodySupport.connectedPort;
		Cell bodyCell = bodyPort.cell;
		if (bodyPort.portNumber != 0
				|| bodyCell.getSymbol() != Cell.SYMBOL_DECODER)
			throw new IllegalStateException();
		DecoderImpl decoder = (DecoderImpl) bodyCell;
		int index = decoder.getIndex();
		if (decoder.getSymbolArity() != 0 || index < 0 || index > 1)
			throw new IllegalStateException();
		return index == 0;
	}

	public long numeralValue(final int[] interactionStat) {
		final boolean TRACE = false;
		if (this.freeVariableCount != 0)
			throw new IllegalStateException("There is free variable(s)");
		PortImpl bodySupport = new LambdaNetImpl(this.bodySupport).bodySupport;
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		PortImpl bodyPort = bodySupport.connectedPort;
		if (bodyPort.portNumber != 0
				|| bodyPort.cell.getSymbol() != Cell.SYMBOL_ABSTRACTOR)
			return -1L;
		bodySupport.connectTo(buildApplication(bodyPort, new DecoderImpl(0)));
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		bodyPort = bodySupport.connectedPort;
		if (TRACE)
			printStat(System.out, bodySupport);
		if (bodyPort.portNumber != 0
				|| bodyPort.cell.getSymbol() != Cell.SYMBOL_ABSTRACTOR)
			return -1L;
		bodySupport.connectTo(buildApplication(bodyPort, new DecoderImpl(1)));
		reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
		bodyPort = bodySupport.connectedPort;
		if (TRACE)
			printStat(System.out, bodySupport);
		CellImpl bodyCell = bodyPort.cell;
		if (bodyPort.portNumber != 0
				|| bodyCell.getSymbol() != Cell.SYMBOL_DECODER)
			return -1L;
		DecoderImpl decoder = (DecoderImpl) bodyCell;
		long n = 0;
		while (decoder.getSymbolArity() == 1) {
			if (decoder.getIndex() != 0)
				return -1L;
			n += decoder.getSequenceCount();
			PortImpl argumentSupport = (PortImpl) decoder.getPort(1);
			bodySupport.connectTo(argumentSupport.connectedPort);
			reduceUntilInterfaceNormalForm(bodySupport, -1, interactionStat);
			bodyPort = bodySupport.connectedPort;
			bodyCell = bodyPort.cell;
			if (bodyPort.portNumber != 0
					|| bodyCell.getSymbol() != Cell.SYMBOL_DECODER)
				return -1L;
			decoder = (DecoderImpl) bodyCell;
		}
		return (decoder.getSymbolArity() == 0 && decoder.getIndex() == 1)
				? (n & Long.MAX_VALUE)
				: -1L;
	}
}