/**
 * Copyright (c) 2005 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda2/).
 */
package org.xmloperator.lambda2.net.model;

/**
 * A cell has a principal port and zero or more auxiliary ports.
 */
public interface Cell {
	public static final int SYMBOL_WIRE = 0;
	public static final int SYMBOL_ERASER = 1;
	public static final int SYMBOL_DUPLICATOR = 2;
	public static final int SYMBOL_CONSTRUCTOR = 3;
	public static final int SYMBOL_ABSTRACTOR = 4;
	public static final int SYMBOL_APPLICATOR = 5;
	public static final int SYMBOL_DECODER = 6;

	public static final char[] SYMBOL_SYMBOLS = {'w', 'e', 'd', 'c', 'a', 'p',
			'x'};

	/**
	 * Returns the Symbol <tt>SYMBOL_*</tt> of this cell.
	 * 
	 * @return the Symbol <tt>SYMBOL_*</tt> of this cell.
	 */
	public int getSymbol();

	/**
	 * Returns the symbol arity of this Cell. It is >= 0.
	 * 
	 * @return the symbol arity of this Cell. It is >= 0.
	 */
	public int getSymbolArity();

	/**
	 * Returns true if this Cell is a WireCell.
	 * 
	 * @return true if this Cell is a WireCell.
	 */
	public boolean isWireCell();

	/**
	 * Returns true if this Cell is an Eraser.
	 * 
	 * @return true if this Cell is an Eraser.
	 */
	public boolean isEraser();

	/**
	 * Returns true if this Cell is a binaryCell, i.e. a constructor or a
	 * duplicator.
	 * 
	 * @return true if this Cell is a binaryCell, i.e. a constructor or a
	 *         duplicator.
	 */
	public boolean isBinaryCell();

	/**
	 * Returns true if this Cell is a PentaryMultiplexor, i.e. an Abstractor or
	 * an Applicator.
	 * 
	 * @return true if this Cell is a PentaryMultiplexor, i.e. an Abstractor or
	 *         an Applicator.
	 */
	public boolean isPentaryMultiplexor();

	/**
	 * Returns true if this Cell is a DecodingCell.
	 * 
	 * @return true if this Cell is a DecodingCell.
	 */
	public boolean isDecoder();

	/**
	 * Returns the principal port of this cell. This an alias for getPort(0).
	 * 
	 * @return the principal port of this cell. This an alias for getPort(0).
	 */
	public Port getPrincipalPort();

	/**
	 * Returns the port with the given number.
	 * 
	 * @param portNumber
	 *            0 for the principal port, 1 to the symbol arity for the
	 *            auxiliary ports.
	 * @return the port with the given number.
	 */
	public Port getPort(int portNumber);

	/**
	 * Adds its ports to a given Stack.
	 * 
	 * @param portStack
	 *            a Stack of ports.
	 */
	public void collectPorts(Stack portStack);

	/**
	 * Clones this Cell and returns the clone.
	 * 
	 * @return this Cell and returns the clone.
	 */
	public Cell cloneCell();

	public static final int INTERACTION_WIRE = 0;
	public static final int INTERACTION_ANNIHILATION_ERASER = 1;
	public static final int INTERACTION_ERASING_DUPLICATOR = 2;
	public static final int INTERACTION_ERASING_CONSTRUCTOR = 3;
	public static final int INTERACTION_ERASING_ABSTRACTOR = 4;
	public static final int INTERACTION_ERASING_APPLICATOR = 5;
	public static final int INTERACTION_ERASING_DECODER = 6;
	public static final int INTERACTION_ANNIHILATION_DUPLICATOR = 7;
	public static final int INTERACTION_ANNIHILATION_CONSTRUCTOR = 8;
	public static final int INTERACTION_CONSTRUCTOR_DUPLICATOR = 9;
	public static final int INTERACTION_CONSTRUCTOR_DUPLICATOR_CONTRUCTORS = 10;
	public static final int INTERACTION_DUPLICATING_ABSTRACTOR = 11;
	public static final int INTERACTION_DUPLICATING_APPLICATOR = 12;
	public static final int INTERACTION_ABSTRACTOR_APPLICATOR = 13;
	public static final int INTERACTION_DUPLICATING_DECODER = 14;
	public static final int INTERACTION_MERGING_DECODER = 15;
	public static final int INTERACTION_MERGING_DECODERS = 16;
	public static final int INTERACTION_APPLICATOR_DECODER = 17;

	public static final String[] INTERACTION_SYMBOLS = {"w-t", "e-e", "e-d",
			"e-c", "e-a", "e-p", "e-x", "d-d", "c-c", "c-d", "c-dcc", "d-a",
			"d-p", "a-p", "d-x", "xd-x", "xxd-x", "p-x"};

	/**
	 * Interacts with the opposite Cell.
	 * <p>
	 * Pre-condition: the two cells must have their principal Ports connected
	 * and have their all auxiliary Ports connected.
	 * </p>
	 * 
	 * @return the interaction type <tt>INTERACTION_*</tt>.
	 * @throws IllegalStateException
	 *             if this Cell is not connected to a principal Port.
	 */
	public int interact();
}
